import json

import pytest

from repoproviders import resolve
from repoproviders.resolvers.serialize import to_json


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com", None),
        (
            "https://github.com/pyOpenSci",
            {
                "certainity": "MaybeExists",
                "kind": "GitHubURL",
                "data": {
                    "installation": "https://github.com",
                    "url": "https://github.com/pyOpenSci",
                },
            },
        ),
        (
            "https://github.com/pyOpenSci/pyos-package-template/tree/main/includes/licenses",
            {
                "certainity": "Exists",
                "kind": "ImmutableGit",
                "data": {
                    "repo": "https://github.com/pyOpenSci/pyos-package-template",
                    "ref": "c77ad6399f713ee3a021ef52b069e56b17de24a7",
                },
            },
        ),
        (
            "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            {
                "certainity": "Exists",
                "kind": "DataverseDataset",
                "data": {
                    "installationUrl": "https://dataverse.harvard.edu",
                    "persistentId": "doi:10.7910/DVN/6ZXAGT",
                },
            },
        ),
        (
            "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777",
            {
                "certainity": "Exists",
                "kind": "ImmutableFigshareDataset",
                "data": {
                    "installation": {
                        "url": "https://figshare.com/",
                        "apiUrl": "https://api.figshare.com/v2/",
                    },
                    "articleId": 9782777,
                    "version": 3,
                },
            },
        ),
        (
            "https://github.com/yuvipanda/does-not-exist-e43",
            {
                "certainity": "DoesNotExist",
                "kind": "ImmutableGit",
                "data": {
                    "kind": "ImmutableGit",
                    "message": "Could not access git repository at https://github.com/yuvipanda/does-not-exist-e43",
                },
            },
        ),
    ),
)
async def test_to_json(url, expected):
    # This also tests to_dict anyway
    answers = await resolve(url, True)
    if expected is None:
        assert answers == []
    else:
        assert answers is not None
        assert to_json(answers[-1]) == json.dumps(expected)



import os

import aiohttp
from yarl import URL

from .base import DoesNotExist, Exists, MaybeExists
from .repos import (
    DataverseDataset,
    DataverseURL,
    Doi,
    FigshareDataset,
    FigshareURL,
    ImmutableFigshareDataset,
    ZenodoDataset,
    ZenodoURL,
)


class DoiResolver:
    """
    A *handle* resolver, called a Doi resolver because that's the most common handle
    """

    async def resolve(self, question: URL) -> Exists[Doi] | DoesNotExist[Doi] | None:
        # Check if this is a valid doi or handle
        if question.scheme in ("doi", "hdl"):
            doi = question.path
        elif question.scheme in ("http", "https") and question.host in (
            "doi.org",
            "www.doi.org",
            "hdl.handle.net",
        ):
            doi = question.path.lstrip("/")
        elif question.scheme == "" and question.path.startswith("10."):
            # Handles in general are defined as <naming-authority>/<handle> (https://datatracker.ietf.org/doc/html/rfc3650#section-3)
            # however, this is far too broad, as even a file path like `hello/world` will satisfy it. Eventually, could find a list
            # of registered handle prefixes to validate the `<naming-authority>` part. In the meantime, we only check for a `10.` prefix,
            # which is for the most popular kind of handle - a DOI.
            # This is only for convenience - in cases when the user pastes in a DOI but doesn't actually say doi:.
            doi = question.path
        else:
            # Not a DOI or handle
            return None

        # TODO: Make the handle resolver we use configurable
        api_url = f"https://doi.org/api/handles/{doi}"

        # FIXME: Reuse this
        async with aiohttp.ClientSession() as session:
            resp = await session.get(api_url)

            if resp.status == 404:
                # This is a validly *formatted* DOI, but it's not actually a dOI
                return DoesNotExist(Doi, f"{doi} is not a registered DOI or handle")
            elif resp.status == 200:
                data = await resp.json()

                # Pick the first URL we find from the doi response
                for v in data["values"]:
                    if v["type"] == "URL":
                        return Exists(Doi(URL(v["data"]["value"])))

                # No URLs found for this DOI, so we treat it as DoesNotExist
                return DoesNotExist(Doi, f"{doi} does not point to any URL")
            else:
                # Some other kind of failure, let's propagate our error up
                resp.raise_for_status()
                # This should not actually be reached, but the explicit return None makes mypy happy
                return None


class DataverseResolver:
    async def get_dataset_id_from_file_id(
        self, installation_url: URL, file_id: str
    ) -> str | None:
        """
        Return the persistent_id (DOI) of a dataset that a given file_id (int or doi) belongs to
        """
        if file_id.isdigit():
            # the file_id is an integer, rather than a persistent id (DOI)
            api_url = installation_url / "api/files" / file_id
            api_url = api_url.with_query(returnDatasetVersion="true")
        else:
            # the file_id is a doi itself
            api_url = installation_url / "api/files/:persistentId"
            api_url = api_url.with_query(
                returnDatasetVersion="true", persistentId=file_id
            )

        async with aiohttp.ClientSession() as session:
            resp = await session.get(api_url)

        if resp.status == 404:
            return None
        else:
            resp.raise_for_status()

        data = (await resp.json())["data"]
        return data["datasetVersion"]["datasetPersistentId"]

    async def resolve(
        self, question: DataverseURL
    ) -> Exists[DataverseDataset] | DoesNotExist[DataverseDataset] | None:
        url = question.url
        path = url.path
        qs = url.query

        # https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/TJCLKP
        if path.startswith("/citation"):
            persistent_id = qs["persistentId"]
            # We don't know if this is a dataset or file id yet
            verified_dataset = False
        # https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/TJCLKP
        elif path.startswith("/dataset.xhtml"):
            #  https://dataverse.harvard.edu/api/access/datafile/3323458
            persistent_id = qs["persistentId"]
            # We haven't verified this dataset exists
            verified_dataset = False
        elif path.startswith("/api/access/datafile"):
            # What we have here is an entity id, which we can use to get a persistentId
            file_id = os.path.basename(path)
            pid_maybe = await self.get_dataset_id_from_file_id(
                question.installation, file_id
            )
            if pid_maybe is None:
                return DoesNotExist(
                    DataverseDataset,
                    f"No file with id {file_id} found in dataverse installation {question.installation}",
                )
            else:
                persistent_id = pid_maybe

            # We know persistent_id is a dataset, because we asked the API!
            verified_dataset = True
        elif path.startswith("/file.xhtml"):
            file_id = qs["persistentId"]
            pid_maybe = await self.get_dataset_id_from_file_id(
                question.installation, file_id
            )
            if pid_maybe is None:
                return DoesNotExist(
                    DataverseDataset,
                    f"No file with id {file_id} found in dataverse installation {question.installation}",
                )
            else:
                persistent_id = pid_maybe
            # We know persistent_id is a dataset, because we asked the API!
            verified_dataset = True
        else:
            # This URL is not actually a dataverse dataset URL
            return None

        if not verified_dataset:
            # citations can be either datasets or files - we don't know. The most common case is that it is
            # a dataset, so we check if it is first.
            api_url = question.installation / "api/datasets/:persistentId"
            api_url = api_url.with_query(persistentId=persistent_id)
            async with aiohttp.ClientSession() as session:
                resp = await session.get(api_url)

            if resp.status == 404:
                # This persistent id is *not* a dataset. Maybe it's a file?
                pid_maybe = await self.get_dataset_id_from_file_id(
                    question.installation, persistent_id
                )
                if pid_maybe is None:
                    # This is not a file either, so this citation doesn't exist
                    return DoesNotExist(
                        DataverseDataset,
                        f"{persistent_id} is neither a file nor a dataset in {question.installation}",
                    )
                else:
                    persistent_id = pid_maybe
            elif resp.status == 200:
                # This *is* a dataset, we just verified it with this query
                verified_dataset = True
            else:
                # Any other errors should propagate
                resp.raise_for_status()

                return None

        return Exists(DataverseDataset(question.installation, persistent_id))


class ZenodoResolver:
    """
    A resolver for datasets hosted on https://inveniosoftware.org/ (such as Zenodo)
    """

    async def resolve(
        self, question: ZenodoURL
    ) -> MaybeExists[ZenodoDataset] | DoesNotExist[ZenodoDataset] | None:
        if not (
            # After the base URL, the URL structure should start with either record or records
            question.url.path[len(question.installation.path) :].startswith("record/")
            or question.url.path[len(question.installation.path) :].startswith(
                "records/"
            )
            or question.url.path[len(question.installation.path) :].startswith("doi/")
        ):
            return None
        # For URLs of form https://zenodo.org/doi/<doi>, the record_id can be resolved by making a
        # HEAD request and following it. This is absolutely *unideal* - you would really instead want
        # to make an API call. But I can't seem to find anything in the REST API that would let me give
        # it a DOI and return a record_id. And these dois can resolve to *different* records over time,
        # so let's actively resolve them here to match the ZenodoDataset's immutable property
        if question.url.path[len(question.installation.path) :].startswith("doi/"):
            url_parts = question.url.path.split("/")
            if len(url_parts) != 4:
                # Not a correctly formatted DOI URL
                return None

            async with aiohttp.ClientSession() as session:
                resp = await session.head(question.url)

            if resp.status == 404:
                return DoesNotExist(
                    ZenodoDataset, f"{question.url} is not a valid Zenodo DOI URL"
                )
            redirect_location = resp.headers["Location"]

            return await self.resolve(
                ZenodoURL(question.installation, URL(redirect_location))
            )
        else:
            # URL is /record or /records
            # Record ID is the last part of the URL path
            return MaybeExists(ZenodoDataset(question.installation, question.url.name))


class FigshareResolver:
    async def resolve(
        self, question: FigshareURL
    ) -> MaybeExists[FigshareDataset] | None:
        # After the base URL, the URL structure should start with either articles or account/articles
        if not (
            question.url.path[len(question.installation.url.path) :].startswith(
                "articles/"
            )
            or question.url.path[len(question.installation.url.path) :].startswith(
                "account/articles/"
            )
        ):
            return None
        # Figshare article IDs are integers, and so are version IDs
        # If last two segments of the URL are integers, treat them as article ID and version ID
        # If not, treat it as article ID only
        parts = question.url.path.split("/")
        if parts[-1].isdigit() and parts[-2].isdigit():
            return MaybeExists(
                FigshareDataset(question.installation, int(parts[-2]), int(parts[-1]))
            )
        elif parts[-1].isdigit():
            return MaybeExists(
                FigshareDataset(question.installation, int(parts[-1]), None)
            )
        else:
            return None


class ImmutableFigshareResolver:
    async def resolve(
        self, question: FigshareDataset
    ) -> (
        Exists[ImmutableFigshareDataset]
        | MaybeExists[ImmutableFigshareDataset]
        | DoesNotExist[ImmutableFigshareDataset]
        | None
    ):
        if question.version is not None:
            # Version already specified, just return
            return MaybeExists(
                ImmutableFigshareDataset(
                    question.installation, question.articleId, question.version
                )
            )

        api_url = (
            question.installation.apiUrl
            / "articles"
            / str(question.articleId)
            / "versions"
        )

        async with aiohttp.ClientSession() as session:
            resp = await session.get(api_url)

        if resp.status == 404:
            return DoesNotExist(
                ImmutableFigshareDataset,
                f"Article ID {question.articleId} not found on figshare installation {question.installation.url}",
            )
        elif resp.status == 200:
            data = await resp.json()
            return Exists(
                ImmutableFigshareDataset(
                    question.installation, question.articleId, data[-1]["version"]
                )
            )
        else:
            # All other status codes should raise an error
            resp.raise_for_status()
            # This should not actually be reached, but the explicit return None makes mypy happy
            return None



import pytest
from yarl import URL

from repoproviders.resolvers.base import Exists
from repoproviders.resolvers.doi import Doi, DoiResolver


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", None),
        # doi schema'd URI
        (
            "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            Exists(
                Doi(
                    URL(
                        "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                    )
                )
            ),
        ),
        # handle schema'd URI
        (
            "hdl:11529/10016",
            Exists(
                Doi(
                    URL(
                        "https://data.cimmyt.org/dataset.xhtml?persistentId=hdl:11529/10016"
                    )
                )
            ),
        ),
        # For convenience, we do accept DOIs without a scheme
        (
            "10.7910/DVN/6ZXAGT/3YRRYJ",
            Exists(
                Doi(
                    URL(
                        "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                    )
                )
            ),
        ),
        # But not handles without a scheme
        ("11529/10016", None),
        # Three DOI resolution URLs
        (
            "https://doi.org/10.7910/DVN/6ZXAGT/3YRRYJ",
            Exists(
                Doi(
                    URL(
                        "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                    )
                )
            ),
        ),
        (
            "https://www.doi.org/10.7910/DVN/6ZXAGT/3YRRYJ",
            Exists(
                Doi(
                    URL(
                        "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                    )
                )
            ),
        ),
        (
            "https://hdl.handle.net/10.7910/DVN/6ZXAGT/3YRRYJ",
            Exists(
                Doi(
                    URL(
                        "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                    )
                )
            ),
        ),
    ),
)
async def test_doi(url, expected):
    doi = DoiResolver()
    assert await doi.resolve(URL(url)) == expected



import json
from pathlib import Path
from typing import Callable

from yarl import URL

from .base import MaybeExists, Repo
from .repos import (
    DataverseURL,
    Doi,
    FigshareInstallation,
    FigshareURL,
    GistURL,
    GitHubURL,
    GitLabURL,
    ZenodoURL,
)


class WellKnownProvidersResolver:
    def detect_github(self, question: URL) -> GitHubURL | None:
        # git+<scheme> urls are handled by a different resolver
        if question.scheme not in ("http", "https") or (
            question.host != "github.com" and question.host != "www.github.com"
        ):
            # TODO: Allow configuring for GitHub enterprise
            return None
        else:
            return GitHubURL(URL("https://github.com"), question)

    def detect_gist(self, question: URL) -> GistURL | None:
        if question.host == "gist.github.com":
            return GistURL(URL("https://gist.github.com"), question)
        else:
            # TODO: Allow configuring for GitHub enterprise
            return None

    def detect_gitlab(self, question: URL) -> GitLabURL | None:
        # git+<scheme> urls are handled by a different resolver
        if question.scheme not in ("http", "https") or (
            question.host != "gitlab.com" and question.host != "www.gitlab.com"
        ):
            # TODO: Allow configuring for GitHub enterprise
            return None
        else:
            return GitLabURL(URL("https://gitlab.com"), question)

    def detect_zenodo(self, question: URL) -> ZenodoURL | None:
        KNOWN_INSTALLATIONS = [
            URL("https://sandbox.zenodo.org/"),
            URL("https://zenodo.org/"),
            URL("https://data.caltech.edu/"),
        ]
        # Intentionally don't check for record/ records/ doi/ etc. That should be all
        # handled by the zenodo resolver
        installation = next(
            (
                installation
                for installation in KNOWN_INSTALLATIONS
                # Intentionally don't check for scheme validity, to support interchangeable http and https URLs
                if installation.host == question.host
                # Check for base URL, to support installations on base URL other than /
                and question.path.startswith(installation.path)
            ),
            None,
        )
        if installation is None:
            return None

        return ZenodoURL(installation, question)

    def detect_dataverse(self, question: URL) -> DataverseURL | None:

        if not hasattr(self, "_dataverse_installation_urls"):
            # Get a list of installation URLs for known dataverse installations
            data_file = Path(__file__).parent / "dataverse.json"
            with open(data_file) as fp:
                installations = json.load(fp)["installations"]

            # Parse all the URLs of installations once, so we can quickly use them for validating URLs passed in
            # TODO: Use a better datastructure here (like a trie?)
            # The structure says 'hostname' but it sometimes has URL prefixes! But no scheme ever
            self._dataverse_installation_urls = [
                URL(f"https://{i['hostname']}") for i in installations
            ]

        # Check if URL is under one of the installation URLs we have.
        installation = next(
            (
                installation
                for installation in self._dataverse_installation_urls
                # Intentionally don't check for scheme validity, to support interchangeable http and https URLs
                if installation.host == question.host
                and question.path.startswith(installation.path)
            ),
            None,
        )
        if installation is None:
            return None
        else:
            return DataverseURL(installation, question)

    def detect_figshare(self, question: URL) -> FigshareURL | None:
        KNOWN_INSTALLATIONS = [
            FigshareInstallation(
                URL("https://figshare.com/"), URL("https://api.figshare.com/v2/")
            )
        ]

        # Check if this is a figshare URL of any sort - *not* that its' a figshare 'article' URL
        # That logic lives within the FigshareResolver itself, so it can be used for other FigshareURLs
        # generated by guessers
        installation = next(
            (
                installation
                for installation in KNOWN_INSTALLATIONS
                # Intentionally don't check for scheme validity, to support interchangeable http and https URLs
                if installation.url.host == question.host
                # Check for base URL, to support installations on base URL other than /
                and question.path.startswith(installation.url.path)
            ),
            None,
        )
        if installation is None:
            return None

        return FigshareURL(installation, question)

    async def resolve(self, question: URL | Doi) -> MaybeExists[Repo] | None:
        # These detectors are *intentionally* not async, as they should not be doing any
        # network calls
        detectors: list[Callable[[URL], Repo | None]] = [
            self.detect_github,
            self.detect_gist,
            self.detect_dataverse,
            self.detect_zenodo,
            self.detect_figshare,
            self.detect_gitlab,
        ]

        match question:
            case URL():
                url = question
            case Doi(doi_url):
                url = doi_url

        for d in detectors:
            maybe_answer = d(url)
            if maybe_answer is not None:
                return MaybeExists(maybe_answer)

        return None



import pytest
from yarl import URL

from repoproviders.resolvers.base import MaybeExists
from repoproviders.resolvers.repos import (
    DataverseURL,
    Doi,
    FigshareInstallation,
    FigshareURL,
    GistURL,
    GitHubURL,
    GitLabURL,
    ZenodoURL,
)
from repoproviders.resolvers.wellknown import WellKnownProvidersResolver


@pytest.mark.parametrize(
    ("question", "expected"),
    (
        ("https://example.com/something", None),
        # Try a raw git repo - this should be only detected by feature detector
        (
            "https://git.kernel.org/pub/scm/virt/kvm/kvm.git/",
            None,
        ),
        # A github URL even if it's not actually a valid repo
        (
            "https://github.com/settings",
            MaybeExists(
                GitHubURL(URL("https://github.com"), URL("https://github.com/settings"))
            ),
        ),
        # A dataverse URL that doesn't exist in our well known set, so should be None
        (
            "https://demo.dataverse.org/dataset.xhtml?persistentId=doi:10.70122/FK2/MBQA9G",
            None,
        ),
        (
            "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/TJCLKP",
            MaybeExists(
                DataverseURL(
                    URL("https://dataverse.harvard.edu"),
                    URL(
                        "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/TJCLKP"
                    ),
                )
            ),
        ),
        (
            Doi(URL("https://zenodo.org/record/3232985")),
            MaybeExists(
                ZenodoURL(
                    URL("https://zenodo.org"),
                    URL("https://zenodo.org/record/3232985"),
                )
            ),
        ),
        (
            "https://zenodo.org/settings",
            MaybeExists(
                ZenodoURL(
                    URL("https://zenodo.org"),
                    URL("https://zenodo.org/settings"),
                )
            ),
        ),
        (
            "https://figshare.com/browse",
            MaybeExists(
                FigshareURL(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    URL("https://figshare.com/browse"),
                )
            ),
        ),
        (
            "https://gitlab.com/browse",
            MaybeExists(
                GitLabURL(URL("https://gitlab.com"), URL("https://gitlab.com/browse"))
            ),
        ),
        (
            "https://gist.github.com/yuvipanda",
            MaybeExists(
                GistURL(
                    URL("https://gist.github.com"),
                    URL("https://gist.github.com/yuvipanda"),
                )
            ),
        ),
        (
            "https://gist.github.com/JakeWharton/5423616",
            MaybeExists(
                GistURL(
                    URL("https://gist.github.com"),
                    URL("https://gist.github.com/JakeWharton/5423616"),
                )
            ),
        ),
    ),
)
async def test_doi(question, expected):
    wk = WellKnownProvidersResolver()
    if isinstance(question, str):
        question = URL(question)
    assert await wk.resolve(question) == expected



import argparse
import asyncio
import sys
from pathlib import Path

from repoproviders.resolvers.base import DoesNotExist, Exists, MaybeExists

from .fetchers import fetch
from .resolvers import resolve


async def main():
    argparser = argparse.ArgumentParser()

    subparsers = argparser.add_subparsers(required=True, dest="command")
    resolve_subparser = subparsers.add_parser("resolve")

    resolve_subparser.add_argument("question", help="What should we try to resolve?")
    resolve_subparser.add_argument(
        "--no-recurse",
        help="Do not recurse, return after first answer",
        action="store_true",
    )

    fetch_subparser = subparsers.add_parser("fetch")
    fetch_subparser.add_argument("question", help="What should we try to fetch?")
    fetch_subparser.add_argument("output_dir", help="Where to output the fetched repo")

    args = argparser.parse_args()

    if args.command == "resolve":
        answers = await resolve(args.question, recursive=not args.no_recurse)

        if answers:
            for a in answers:
                print(a)
        else:
            print(f"Unable to resolve {args.question}")
    elif args.command == "fetch":
        output_dir = Path(args.output_dir)
        # output_dir must not exist, or be an empty directory
        if output_dir.exists():
            if output_dir.is_file():
                print(
                    f"{output_dir} should either not exist, or be an empty directory. Is a file",
                    file=sys.stderr,
                )
                sys.exit(1)
            else:
                if any(output_dir.iterdir()):
                    print(
                        f"{output_dir} should either not exist, or be an empty directory. Is a non-empty directory",
                        file=sys.stderr,
                    )
                    sys.exit(1)
        else:
            output_dir.mkdir(parents=True)

        answers = await resolve(args.question, recursive=True)
        if answers:
            last_answer = answers[-1]
            match last_answer:
                case Exists(repo) | MaybeExists(repo):
                    await fetch(repo, Path(args.output_dir))
                case DoesNotExist(kind, message):
                    print(
                        f"{args.question} detected to be of kind {kind.__name__} but does not exist: {message}",
                        file=sys.stderr,
                    )
                    sys.exit(1)
        else:
            print(f"Unable to resolve {args.question}")


def cli_entrypoint():
    asyncio.run(main())



import pytest
from yarl import URL

from repoproviders.resolvers.base import DoesNotExist, Exists
from repoproviders.resolvers.doi import DataverseResolver
from repoproviders.resolvers.repos import DataverseDataset, DataverseURL


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # A dataset citation returns the dataset correctly
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/TJCLKP"
                ),
            ),
            Exists(
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/TJCLKP"
                )
            ),
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/TJCLKP"
                ),
            ),
            Exists(
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/TJCLKP"
                )
            ),
        ),
        # Asking for specific files should give us the whole dataset they are a part of
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL("https://dataverse.harvard.edu/api/access/datafile/3323458"),
            ),
            Exists(
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/3MJ7IR"
                )
            ),
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                ),
            ),
            Exists(
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
                )
            ),
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                ),
            ),
            Exists(
                DataverseDataset(
                    URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
                )
            ),
        ),
        # Asking for datasets that don't exist should return DoesNotExist
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/not-found"
                ),
            ),
            DoesNotExist(
                DataverseDataset,
                "doi:10.7910/not-found is neither a file nor a dataset in https://dataverse.harvard.edu",
            ),
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/not-found"
                ),
            ),
            DoesNotExist(
                DataverseDataset,
                "doi:10.7910/not-found is neither a file nor a dataset in https://dataverse.harvard.edu",
            ),
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL("https://dataverse.harvard.edu/api/access/datafile/0"),
            ),
            DoesNotExist(
                DataverseDataset,
                "No file with id 0 found in dataverse installation https://dataverse.harvard.edu",
            ),
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL("https://dataverse.harvard.edu/blaaaah"),
            ),
            None,
        ),
        (
            DataverseURL(
                URL("https://dataverse.harvard.edu"),
                URL(
                    "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/not-found"
                ),
            ),
            DoesNotExist(
                DataverseDataset,
                "No file with id doi:10.7910/not-found found in dataverse installation https://dataverse.harvard.edu",
            ),
        ),
    ),
)
async def test_dataverse(url, expected):
    dv = DataverseResolver()
    assert await dv.resolve(url) == expected



import os
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

import aiohttp

from ..resolvers.doi import ZenodoDataset
from ..utils import download_file


class ZenodoFetcher:
    async def fetch(self, repo: ZenodoDataset, output_dir: Path):
        files_url = repo.installationUrl / "api/records" / repo.recordId / "files"

        async with aiohttp.ClientSession() as session:
            resp = await session.get(files_url)

            # FIXME: Do we assume this persistent_id has been *verified* to exist?
            # What kind of guarantee can our resolver actually give us? hmm
            resp.raise_for_status()

            data = await resp.json()

            # Handle case when we only have one entry, and it's a zip file
            if len(data["entries"]) == 1:
                # Only do this if mimetype is zip
                entry = data["entries"][0]
                if entry["mimetype"] == "application/zip":
                    download_url = entry["links"]["content"]

                    with NamedTemporaryFile() as temp_file:
                        await download_file(session, download_url, Path(temp_file.name))
                        compressed_file = ZipFile(temp_file.name)
                        compressed_file.extractall(output_dir)

                        # If there's only one subdirectory, move its contents to the output directory
                        subdirs = list(output_dir.iterdir())
                        if len(subdirs) == 1:
                            # Just recursively move everything inside the subdir, rather than copy them
                            # This is safer (and probably faster) than using shutil.copytree
                            for d in subdirs[0].iterdir():
                                shutil.move(d, output_dir)
                            os.rmdir(subdirs[0])
                        # We're all done, no more processing to do
                        return

            # For cases with more than one entry, or if the one entry isn't a zip file
            for entry in data["entries"]:
                file_download_url = entry["links"]["content"]

                # FIXME: Handle path traverseal attacks here?
                file_name = entry["key"]

                # Zenodo doesn't support directory structures,
                # so we don't need to handle nesting https://support.zenodo.org/help/en-gb/1-upload-deposit/74-can-i-upload-folders-directories
                await download_file(session, file_download_url, output_dir / file_name)



import dataclasses
import json
from typing import Any

from yarl import URL

from repoproviders.resolvers.base import DoesNotExist, Exists, MaybeExists

from .base import Repo


class JSONEncoder(json.JSONEncoder):
    def default(self, o: URL | Any | type) -> str | Any:
        if isinstance(o, URL):
            return str(o)
        elif isinstance(o, type):
            return o.__name__
        else:
            return super().default(o)


def to_json(answer: DoesNotExist[Repo] | Exists[Repo] | MaybeExists[Repo]):
    """
    Convert an answer into a canonical JSON representation
    """
    return json.dumps(to_dict(answer), cls=JSONEncoder)


def to_dict(
    answer: DoesNotExist[Repo] | Exists[Repo] | MaybeExists[Repo],
) -> dict[str, Any]:
    """
    Convert an answer into a canonical dict representation
    """
    match answer:
        case DoesNotExist(_, _):
            return {
                "certainity": answer.__class__.__name__,
                "kind": answer.kind.__name__,
                "data": dataclasses.asdict(answer),
            }
        case Exists(repo) | MaybeExists(repo):
            return {
                "certainity": answer.__class__.__name__,
                "kind": answer.repo.__class__.__name__,
                "data": dataclasses.asdict(repo),
            }



import asyncio
import secrets
import socket
import sys
import tempfile
from pathlib import Path

import aiohttp
from yarl import URL

from repoproviders.utils import download_file


def random_port() -> int:
    """
    Get a single random port that is *probably* unused
    """
    sock = socket.socket()
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


async def test_download_file():
    port = random_port()
    with tempfile.TemporaryDirectory() as src, tempfile.TemporaryDirectory() as dest:
        test_file = Path(src) / secrets.token_hex(8)
        test_contents = secrets.token_hex(8)
        proc = await asyncio.create_subprocess_exec(
            sys.executable, "-m", "http.server", str(port), "-d", src, "-b", "127.0.0.1"
        )
        try:
            # FIXME: Do this a little more dynamically?
            # Wait for the HTTP server to come up
            await asyncio.sleep(1)

            test_file.write_text(test_contents)

            # Create a nested subdirectory
            dest_file = Path(dest) / secrets.token_hex(8) / secrets.token_hex(8)
            async with aiohttp.ClientSession() as session:
                await download_file(
                    session, URL(f"http://127.0.0.1:{port}/{test_file.name}"), dest_file
                )

            assert dest_file.exists()
            assert dest_file.read_text() == test_contents
        finally:
            proc.kill()
            await proc.wait()



import os
from pathlib import Path

import aiohttp

from ..resolvers.doi import DataverseDataset
from ..utils import download_file


class DataverseFetcher:
    async def fetch(self, repo: DataverseDataset, output_dir: Path):
        files_url = repo.installationUrl / "api/datasets/:persistentId"
        files_url = files_url.with_query(persistentId=repo.persistentId)

        async with aiohttp.ClientSession() as session:
            resp = await session.get(files_url)

            # FIXME: Do we assume this persistent_id has been *verified* to exist?
            # What kind of guarantee can our resolver actually give us? hmm
            resp.raise_for_status()

            files = (await resp.json())["data"]["latestVersion"]["files"]

            for f in files:
                file_download_url = (
                    repo.installationUrl / f"api/access/datafile/{f['dataFile']['id']}"
                )
                file_download_url = file_download_url.with_query(format="original")

                file_name = f["dataFile"].get("originalFileName", f["label"])

                file_path = Path(os.path.join(f.get("directoryLabel", ""), file_name))

                await download_file(session, file_download_url, output_dir / file_path)



import pytest
from yarl import URL

from repoproviders.resolvers.base import MaybeExists
from repoproviders.resolvers.git import Git, GitUrlResolver


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # Not a git url
        ("https://example.com/something", None),
        # Not a real repo, but looks like one
        (
            "git+https://example.com/something",
            MaybeExists(Git("https://example.com/something", "HEAD")),
        ),
        (
            "git+ssh://example.com/something",
            MaybeExists(Git("ssh://example.com/something", "HEAD")),
        ),
        # With a ref
        (
            "git+https://github.com/yuvipanda/requirements@main",
            MaybeExists(Git("https://github.com/yuvipanda/requirements", "main")),
        ),
        # With edge case refs
        (
            "git+https://github.com/yuvipanda/requirements@tag/something",
            MaybeExists(
                Git("https://github.com/yuvipanda/requirements", "tag/something")
            ),
        ),
        (
            "git+https://yuvipanda@github.com/yuvipanda/requirements@tag/something",
            MaybeExists(
                Git(
                    "https://yuvipanda@github.com/yuvipanda/requirements",
                    "tag/something",
                )
            ),
        ),
        (
            "git+https://yuvipanda@github.com/yuvipanda/requirements@tag@something",
            MaybeExists(
                Git(
                    "https://yuvipanda@github.com/yuvipanda/requirements",
                    "tag@something",
                )
            ),
        ),
    ),
)
async def test_giturl(url, expected):
    gu = GitUrlResolver()
    assert await gu.resolve(URL(url)) == expected



from dataclasses import dataclass

from yarl import URL


@dataclass(frozen=True)
class Git:
    repo: str
    ref: str

    immutable = False


@dataclass(frozen=True)
class ImmutableGit:
    """
    Same as Git, but marked to be fully resolved. This implies:

    1. The repository exists, and can be contacted
    2. If ref was a branch or tag, it has been resolved into an immutable commit sha
    3. If ref *looks* like a sha, we assume it exists (without testing it)
    """

    repo: str
    ref: str

    immutable = True


@dataclass(frozen=True)
class GitHubURL:
    """
    A GitHub URL of any sort.

    Not just a repository URL.
    """

    installation: URL
    url: URL

    # URLs can point to whatever
    immutable = False


@dataclass(frozen=True)
class GitHubPR:
    """
    A GitHub Pull Request
    """

    installation: URL
    url: URL

    # PRs can change to whatever
    immutable = False


@dataclass(frozen=True)
class GistURL:
    """
    A Gist URL of any sort.

    Not just cloneable repo URL
    """

    installation: URL
    url: URL

    # URL can point to whatever
    immutable = False


@dataclass(frozen=True)
class GitLabURL:
    """
    A GitLab URL of any sort.

    Not just a repository URL.
    """

    installation: URL
    url: URL

    # URLs can point to whatever
    immutable = False


@dataclass(frozen=True)
class Doi:
    url: URL

    # This needs further investigation
    immutable = False


@dataclass(frozen=True)
class DataverseURL:
    """
    Any kind of URL in any dataverse installation

    Not just for datasets.
    """

    installation: URL
    url: URL

    immutable = False


@dataclass(frozen=True)
class DataverseDataset:
    installationUrl: URL
    persistentId: str

    # Dataverse Datasets also have versions, which are not represented here.
    immutable = False


@dataclass(frozen=True)
class ZenodoURL:
    """
    Any kind of URL in any Zenodo / Invenio installation.

    Not just for records.
    """

    installation: URL
    url: URL

    immutable = False


@dataclass(frozen=True)
class ZenodoDataset:
    installationUrl: URL
    recordId: str

    # Zenodo records are immutable: https://help.zenodo.org/docs/deposit/about-records/#life-cycle
    # When a new version is published, it gets its own record id!
    immutable = True


@dataclass(frozen=True)
class FigshareInstallation:
    url: URL
    apiUrl: URL


@dataclass(frozen=True)
class FigshareURL:
    """
    Any kind of URL in any Figshare installation.

    Not just for articles / datasets.
    """

    installation: FigshareInstallation
    url: URL

    immutable = False


@dataclass(frozen=True)
class FigshareDataset:
    installation: FigshareInstallation
    articleId: int
    version: int | None

    # Figshare articles have versions, and here we don't know if this one does or not
    immutable = False


@dataclass(frozen=True)
class ImmutableFigshareDataset:
    installation: FigshareInstallation
    articleId: int
    # version will always be present when immutable
    version: int

    # We *know* there's a version here
    immutable = True



from dataclasses import Field, dataclass
from typing import Any, ClassVar, Protocol, runtime_checkable


@dataclass(frozen=True)
class DoesNotExist[T: Repo]:
    """
    Resolver recognizes this question, but while resolving determined it does not exist
    """

    # womp womp, we can't really retrieve the value of T at runtime so gotta still pass
    # that in.
    # FIXME: See if we can enforce somehow that kind is also a `Repo`
    kind: type
    message: str


@dataclass(frozen=True)
class Exists[T: Repo]:
    repo: T


@dataclass(frozen=True)
class MaybeExists[T: Repo]:
    repo: T


@runtime_checkable
class Repo(Protocol):
    """
    Represents a Repository
    """

    # Set to true if the repo is identified to be immutable. This usually means one of:
    #
    # 1. The identifier used to refer to the repo is guaranteed to not change (like with zenodo record ids)
    # 2. A version identifier is included as part of the definition (like with Figshare)
    # 3. A content identifiable hash is included as part of the definition (like with ImmutableGit)
    immutable: bool

    # Must also be a dataclass
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


class SupportsResolve(Protocol):
    async def resolve(
        self, question: Any
    ) -> Exists | DoesNotExist | MaybeExists | None:
        pass



# Copyright (c) 2024 Yuvi Panda
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from .fetchers.fetcher import fetch  # noqa: F401
from .resolvers.resolver import resolve  # noqa: F401



import pytest
from yarl import URL

from repoproviders.resolvers.base import MaybeExists
from repoproviders.resolvers.git import GistResolver, Git
from repoproviders.resolvers.repos import GistURL


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # Gist URLs that aren't gists
        (
            GistURL(
                URL("https://gist.github.com"), URL("https://gist.github.com/yuvipanda")
            ),
            None,
        ),
        # We don't support revisions yet
        (
            GistURL(
                URL("https://gist.github.com"),
                URL("https://gist.github.com/JakeWharton/5423616/revisions"),
            ),
            None,
        ),
        # An actual gist
        (
            GistURL(
                URL("https://gist.github.com"),
                URL("https://gist.github.com/JakeWharton/5423616"),
            ),
            MaybeExists(
                repo=Git(
                    repo="https://gist.github.com/JakeWharton/5423616",
                    ref="HEAD",
                )
            ),
        ),
    ),
)
async def test_gist(url, expected):
    gh = GistResolver()
    assert await gh.resolve(url) == expected



import inspect
import types
import typing
from typing import Any

from yarl import URL

from repoproviders.resolvers.feature_detect import FeatureDetectResolver

from .base import DoesNotExist, Exists, MaybeExists, SupportsResolve
from .doi import (
    DataverseResolver,
    DoiResolver,
    FigshareResolver,
    ImmutableFigshareResolver,
    ZenodoResolver,
)
from .git import (
    GistResolver,
    GitHubPRResolver,
    GitHubResolver,
    GitLabResolver,
    GitUrlResolver,
    ImmutableGitResolver,
)
from .wellknown import WellKnownProvidersResolver

ALL_RESOLVERS: list[SupportsResolve] = [
    WellKnownProvidersResolver(),
    GitHubResolver(),
    GitHubPRResolver(),
    GistResolver(),
    GitUrlResolver(),
    GitLabResolver(),
    DoiResolver(),
    ZenodoResolver(),
    FigshareResolver(),
    ImmutableFigshareResolver(),
    DataverseResolver(),
    ImmutableGitResolver(),
    FeatureDetectResolver(),
]

RESOLVER_BY_TYPE: dict[type, list[SupportsResolve]] = {}
for R in ALL_RESOLVERS:
    annotations = inspect.get_annotations(R.resolve)
    supported_types = annotations["question"]
    if isinstance(supported_types, type):
        # Only supports a single type
        RESOLVER_BY_TYPE.setdefault(supported_types, []).append(R)
    elif isinstance(supported_types, types.UnionType):
        for t in typing.get_args(supported_types):
            RESOLVER_BY_TYPE.setdefault(t, []).append(R)


async def resolve(
    question: str | Any, recursive: bool
) -> list[Exists | MaybeExists | DoesNotExist] | None:
    if isinstance(question, str):
        question = URL(question)

    answers: list[Exists | MaybeExists | DoesNotExist] = []
    resp = None

    while True:
        # Get a list of applicable resolvers we can try
        applicable_resolvers = RESOLVER_BY_TYPE.get(type(question), [])

        if not applicable_resolvers:
            # No applicable resolvers found for this question type, we are done
            break

        for r in applicable_resolvers:
            resp = await r.resolve(question)
            if resp is not None:
                # We found an answer!
                answers.append(resp)
                # Break out of the for after we find an answer in each round
                break

        if recursive:
            # If we want a recursive answer, we have to continue iterating
            match resp:
                case DoesNotExist():
                    # Some resolver detected this but we have confirmed it does not exist
                    break
                case Exists(repo):
                    # TODO: Should an "Exists" be further resolved, or does it always indicate an end?
                    question = repo
                case MaybeExists(repo):
                    question = repo
                case None:
                    # No answer was found this round so we are done
                    break
                # We *did* find an answer this round, so we should continue and see if we find more
            resp = None
        else:
            # We are not recursive, so we are done after 1 round regardless
            break

    return answers



import hashlib
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from repoproviders.fetchers import fetch
from repoproviders.resolvers import resolve
from repoproviders.resolvers.base import DoesNotExist


@pytest.mark.parametrize(
    ("questions", "md5tree"),
    [
        (
            # Test fetching a record with a single zip file that we then extract,
            # Also accounting for data.caltech.edu sending back 200 response with Location headers
            ("https://data.caltech.edu/records/996aw-mf266",),
            {
                "CaltechDATA_Usage_Graphs.ipynb": "3bb3a97b879112b1ab70923636d63e87",
                "LICENSE": "be9d12a5904d8e4ef6df52c86d2db34a",
                "requirements.txt": "b0081730d38ec5f28f3fea4f843816ce",
                "README.md": "59a0e5157faef752532fe51a2d490c9c",
                ".gitignore": "2a9ac83919f923cc25ea380b19a2a7d9",
                "codemeta.json": "8aa017724932fbdff3d2240c932487b7",
            },
        ),
        (
            # Test fetching a record with a single zip file, without the 200 / Location header
            # issue
            ("https://sandbox.zenodo.org/records/432153",),
            {
                "LICENSE": "65d3616852dbf7b1a6d4b53b00626032",
                "README.md": "041917c62158f2dec74eb1ead07662f1",
                "rutgers.txt": "2f501f69915cbcf0fb185f9e8bdb1c96",
            },
        ),
        (
            # Test fetching a record with a single file that is *not* a zip file
            ("https://sandbox.zenodo.org/records/415845",),
            {"136_poster_-_Dalia_Al-Shahrabi.pdf": "663a007fead7da7f9f6f7ddae71db254"},
        ),
        (
            # Test fetching multiple files without a zip file
            ("https://sandbox.zenodo.org/records/98954",),
            {
                "sampleFile.json": "9c9cf2a3740a65cc3268e12567eed67b",
                "sampleFile.txt": "fedb53c2017c1aad5bf9293b7ce03a71",
            },
        ),
    ],
)
async def test_fetch(questions: list[str], md5tree: dict[str, str]):
    for question in questions:
        with TemporaryDirectory() as d:
            output_dir = Path(d)
            answers = await resolve(question, True)

            assert answers is not None
            assert not isinstance(answers[-1], DoesNotExist)

            await fetch(answers[-1].repo, output_dir)

            # Verify md5 sum of the files we expect to find
            for subpath, expected_sha in md5tree.items():
                with open(output_dir / subpath, "rb") as f:
                    h = hashlib.md5()
                    h.update(f.read())
                    assert h.hexdigest() == expected_sha



from json import JSONDecodeError

import aiohttp
from yarl import URL

from .base import Exists, MaybeExists
from .repos import DataverseURL, Git, GitLabURL


class FeatureDetectResolver:
    """
    Use external network calls to detect what kinda URL this is.

    Resolver of last resort!
    """

    async def is_git_repo(
        self, session: aiohttp.ClientSession, url: URL
    ) -> Exists[Git] | None:
        """
        Return true if url is a git repository that supports the smart HTTP git protocol
        """
        # Read https://github.com/git/git/blob/master/Documentation/gitprotocol-http.txt
        # to understand the smart HTTP git protocol better.
        # Short version is that we look for a 200 response in $GIT_URL/info/refs?service=git-upload-pack
        # To determine if this is a git repo
        refs_url = url / "info/refs"
        refs_url = refs_url.with_query(service="git-upload-pack")

        resp = await session.get(refs_url)

        if resp.status == 200:
            return Exists(Git(str(url), "HEAD"))

        # Not a smart git URL
        return None

    async def is_dataverse(
        self, session: aiohttp.ClientSession, url: URL
    ) -> MaybeExists[DataverseURL] | None:
        """
        Check if a given URL is under a dataverse install
        """

        # Make an API call to check if this is a dataverse instance
        # https://guides.dataverse.org/en/latest/api/native-api.html#show-dataverse-software-version-and-build-number
        # FIXME: This assumes that the dataverse instance is hosted at the root of the server,
        # without any other path prefix
        installation = url.with_path("/").with_fragment(None).with_query(None)
        api_url = installation.with_path("/api/info/version")
        resp = await session.get(api_url)

        if resp.status == 200:
            try:
                version_data = await resp.json()
                if version_data.get("status") == "OK" and "version" in version_data.get(
                    "data", {}
                ):
                    return MaybeExists(DataverseURL(installation, url))
            except:
                pass

        return None

    async def is_gitlab(
        self, session: aiohttp.ClientSession, question: URL
    ) -> MaybeExists[GitLabURL] | None:
        # A lot of GitLab APIs seem to require auth to hit, including the version API
        # So instead, we hit the OpenID Connect Well Known Endpoint (https://docs.gitlab.com/ee/integration/openid_connect_provider.html#settings-discovery)
        # And check for GitLab specific supported claims.
        installation = question.with_path("/").with_query(None).with_fragment(None)
        openid_config_url = installation / ".well-known/openid-configuration"

        resp = await session.get(openid_config_url)

        if resp.status != 200:
            return None

        try:
            data = await resp.json()
        except JSONDecodeError:
            return None

        if "https://gitlab.org/claims/groups/owner" in data.get("claims_supported", {}):
            return MaybeExists(GitLabURL(installation, question))
        else:
            return None

    async def resolve(
        self, question: URL
    ) -> Exists[Git] | MaybeExists[DataverseURL] | MaybeExists[GitLabURL] | None:
        if question.scheme not in ("http", "https"):
            return None

        detectors = (self.is_dataverse, self.is_gitlab, self.is_git_repo)

        async with aiohttp.ClientSession() as session:
            for g in detectors:
                maybe_answer = await g(session, question)
                if maybe_answer is not None:
                    return maybe_answer

        return None



import pytest
from yarl import URL

from repoproviders.resolvers.base import MaybeExists
from repoproviders.resolvers.git import Git, GitLabResolver
from repoproviders.resolvers.repos import GitLabURL


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # GitLab URLs that are not repos
        (
            GitLabURL(URL("https://gitlab.com"), URL("https://gitlab.com/mosaik/")),
            None,
        ),
        (
            GitLabURL(
                URL("https://gitlab.com"),
                URL("https://gitlab.com/groups/mosaik/-/packages"),
            ),
            None,
        ),
        (
            GitLabURL(
                URL("https://gitlab.com"),
                # Put a - at the end to test for edge cases
                URL("https://gitlab.com/groups/mosaik/something/something/-"),
            ),
            None,
        ),
        (
            GitLabURL(
                URL("https://gitlab.com"),
                URL("https://gitlab.com/mosaik/mosaik/-/merge_requests/194"),
            ),
            None,
        ),
        # Simple GitLab repo URL
        (
            GitLabURL(
                URL("https://gitlab.com"), URL("https://gitlab.com/mosaik/mosaik")
            ),
            MaybeExists(Git("https://gitlab.com/mosaik/mosaik", "HEAD")),
        ),
        # blobs and tree
        (
            GitLabURL(
                URL("https://gitlab.com"),
                URL(
                    "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder/-/blob/b2c44e7f804dc1634681540582b1731e0393f69b/03_Same_Time_Loop/_01_controller_master.ipynb?ref_type=heads"
                ),
            ),
            MaybeExists(
                Git(
                    "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder",
                    "b2c44e7f804dc1634681540582b1731e0393f69b",
                )
            ),
        ),
        (
            GitLabURL(
                URL("https://gitlab.com"),
                URL(
                    "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder/-/tree/b2c44e7f804dc1634681540582b1731e0393f69b/03_Same_Time_Loop?ref_type=heads"
                ),
            ),
            MaybeExists(
                Git(
                    "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder",
                    "b2c44e7f804dc1634681540582b1731e0393f69b",
                )
            ),
        ),
        (
            GitLabURL(
                URL("https://gitlab.wikimedia.org"),
                URL(
                    "https://gitlab.wikimedia.org/toolforge-repos/toolviews/-/tree/main/toolviews?ref_type=heads"
                ),
            ),
            MaybeExists(
                Git("https://gitlab.wikimedia.org/toolforge-repos/toolviews", "main")
            ),
        ),
        (
            GitLabURL(
                URL("https://gitlab.wikimedia.org"),
                URL(
                    "https://gitlab.wikimedia.org/toolforge-repos/toolviews/-/blob/main/toolviews/__init__.py?ref_type=heads"
                ),
            ),
            MaybeExists(
                Git("https://gitlab.wikimedia.org/toolforge-repos/toolviews", "main")
            ),
        ),
        (
            GitLabURL(
                URL("https://gitlab.com"),
                URL(
                    "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder/-/tree/b2c44e7f804dc1634681540582b1731e0393f69b/03_Same_Time_Loop?ref_type=heads"
                ),
            ),
            MaybeExists(
                Git(
                    "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder",
                    "b2c44e7f804dc1634681540582b1731e0393f69b",
                )
            ),
        ),
        (
            GitLabURL(
                URL("https://gitlab.com"),
                URL("https://gitlab.com/yuvipanda/does-not-exist-e43"),
            ),
            MaybeExists(
                Git(repo="https://gitlab.com/yuvipanda/does-not-exist-e43", ref="HEAD")
            ),
        ),
        # Non repo URLs should simply be detected to be not a repo
        (
            GitLabURL(
                URL("https://gitlab.wikimedia.org"),
                URL(
                    "https://gitlab.wikimedia.org/toolforge-repos/toolviews/-/merge_requests/11"
                ),
            ),
            None,
        ),
    ),
)
async def test_gitlab(url, expected):
    gl = GitLabResolver()
    assert await gl.resolve(url) == expected



import pytest
from yarl import URL

from repoproviders.resolvers.base import Exists, MaybeExists
from repoproviders.resolvers.feature_detect import FeatureDetectResolver
from repoproviders.resolvers.git import Git
from repoproviders.resolvers.repos import DataverseURL


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", None),
        # Try a raw git repo
        (
            "https://git.kernel.org/pub/scm/virt/kvm/kvm.git/",
            Exists(Git("https://git.kernel.org/pub/scm/virt/kvm/kvm.git/", "HEAD")),
        ),
        # A Codeberg repo
        (
            "https://codeberg.org/Codeberg/Documentation",
            Exists(Git("https://codeberg.org/Codeberg/Documentation", "HEAD")),
        ),
        # A dataverse URL from a dataverse installation that is *not* in our well known list
        (
            "https://demo.dataverse.org/dataset.xhtml?persistentId=doi:10.70122/FK2/MBQA9G",
            MaybeExists(
                DataverseURL(
                    URL("https://demo.dataverse.org"),
                    URL(
                        "https://demo.dataverse.org/dataset.xhtml?persistentId=doi:10.70122/FK2/MBQA9G"
                    ),
                )
            ),
        ),
        # A dataverse URL that does exist in our list, but good to make sure guess works with it still
        (
            "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/TJCLKP",
            MaybeExists(
                DataverseURL(
                    URL("https://dataverse.harvard.edu/"),
                    URL(
                        "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/TJCLKP"
                    ),
                )
            ),
        ),
    ),
)
async def test_doi(url, expected):
    fd = FeatureDetectResolver()
    assert await fd.resolve(URL(url)) == expected



import inspect
import types
import typing
from pathlib import Path
from typing import Any

from repoproviders.fetchers.zenodo import ZenodoFetcher

from .base import SupportsFetch
from .dataverse import DataverseFetcher
from .git import ImmutableGitFetcher

ALL_FETCHERS: list[SupportsFetch] = [
    ImmutableGitFetcher(),
    DataverseFetcher(),
    ZenodoFetcher(),
]

FETCHER_BY_TYPE: dict[type, SupportsFetch] = {}
for R in ALL_FETCHERS:
    annotations = inspect.get_annotations(R.fetch)
    supported_types = annotations["repo"]
    if isinstance(supported_types, type):
        # Only supports a single type
        FETCHER_BY_TYPE[supported_types] = R
    elif isinstance(supported_types, types.UnionType):
        for t in typing.get_args(supported_types):
            FETCHER_BY_TYPE[t] = R


async def fetch(question: Any, output_dir: Path):
    fetcher = FETCHER_BY_TYPE[type(question)]

    await fetcher.fetch(question, output_dir)



import asyncio
from pathlib import Path
from subprocess import CalledProcessError

from ..resolvers.git import ImmutableGit


class ImmutableGitFetcher:
    async def fetch(self, repo: ImmutableGit, output_dir: Path):

        # Assume output_dir is empty
        command = [
            "git",
            "clone",
            "--filter=tree:0",
            "--recurse-submodules",
            repo.repo,
            str(output_dir),
        ]
        proc = await asyncio.create_subprocess_exec(
            *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = [s.decode().strip() for s in await proc.communicate()]
        retcode = await proc.wait()

        if retcode != 0:
            # FIXME: Raise a more helpful error?
            raise CalledProcessError(retcode, command, stdout, stderr)

        command = ["git", "checkout", repo.ref]

        proc = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=str(output_dir)
        )

        stdout, stderr = [s.decode().strip() for s in await proc.communicate()]
        retcode = await proc.wait()

        if retcode != 0:
            # FIXME: Raise a more helpful error?
            raise CalledProcessError(retcode, command, stdout, stderr)



from pathlib import Path

import aiohttp
from yarl import URL


async def download_file(session: aiohttp.ClientSession, url: URL, output_path: Path):
    # Read in 4k chunks
    CHUNK_SIZE = 4 * 1024
    resp = await session.get(url)

    if resp.status == 200 and "Location" in resp.headers:
        # Some providers (lookin at you, data.caltech.edu) send a Location header
        # *but with a 200 status code*. This is invalid and bogus, yet we have to
        # honor it. Sigh
        return await download_file(session, URL(resp.headers["Location"]), output_path)

    resp.raise_for_status()

    if not output_path.parent.exists():
        # Leave the exist_ok=True here despite the check, to handle possible
        # race conditions in the future if we ever parallelize this
        output_path.parent.mkdir(parents=True, exist_ok=True)

    with open(output_path, "wb") as f:
        async for chunk in resp.content.iter_chunked(CHUNK_SIZE):
            f.write(chunk)



import pytest
from yarl import URL

from repoproviders.resolvers.base import DoesNotExist, Exists, MaybeExists
from repoproviders.resolvers.doi import FigshareResolver, ImmutableFigshareResolver
from repoproviders.resolvers.repos import (
    FigshareDataset,
    FigshareInstallation,
    FigshareURL,
    ImmutableFigshareDataset,
)


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # A non-dataset URL
        (
            FigshareURL(
                FigshareInstallation(
                    URL("https://figshare.com/"),
                    URL("https://api.figshare.com/v2/"),
                ),
                URL("https://figshare.com/browse"),
            ),
            None,
        ),
        # A non-dataset URL that looks suspiciously like a dataset URL
        (
            FigshareURL(
                FigshareInstallation(
                    URL("https://figshare.com/"),
                    URL("https://api.figshare.com/v2/"),
                ),
                URL(
                    "https://figshare.com/collections/Risk_reduction_in_SARS-CoV-2_infection_and_reinfection_conferred_by_humoral_antibody_levels_among_essential_workers_during_Omicron_predominance/7605487",
                ),
            ),
            None,
        ),
        # Some old school URLs
        (
            FigshareURL(
                FigshareInstallation(
                    URL("https://figshare.com/"),
                    URL("https://api.figshare.com/v2/"),
                ),
                URL(
                    "https://figshare.com/articles/title/9782777",
                ),
            ),
            MaybeExists(
                FigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    None,
                )
            ),
        ),
        (
            FigshareURL(
                FigshareInstallation(
                    URL("https://figshare.com/"),
                    URL("https://api.figshare.com/v2/"),
                ),
                URL(
                    "https://figshare.com/articles/title/9782777/2",
                ),
            ),
            MaybeExists(
                FigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    2,
                )
            ),
        ),
        # New style URLs
        (
            FigshareURL(
                FigshareInstallation(
                    URL("https://figshare.com/"),
                    URL("https://api.figshare.com/v2/"),
                ),
                URL(
                    "https://figshare.com/articles/code/Binder-ready_openSenseMap_Analysis/9782777",
                ),
            ),
            MaybeExists(
                FigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    None,
                )
            ),
        ),
        (
            FigshareURL(
                FigshareInstallation(
                    URL("https://figshare.com/"),
                    URL("https://api.figshare.com/v2/"),
                ),
                URL(
                    "https://figshare.com/articles/code/Binder-ready_openSenseMap_Analysis/9782777/3",
                ),
            ),
            MaybeExists(
                FigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    3,
                )
            ),
        ),
    ),
)
async def test_figshare(url, expected):
    fs = FigshareResolver()
    assert await fs.resolve(url) == expected


@pytest.mark.parametrize(
    ("question", "expected"),
    (
        (
            FigshareDataset(
                FigshareInstallation(
                    URL("https://figshare.com/"), URL("https://api.figshare.com/v2/")
                ),
                9782777,
                None,
            ),
            Exists(
                ImmutableFigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    3,
                )
            ),
        ),
        (
            FigshareDataset(
                FigshareInstallation(
                    URL("https://figshare.com/"), URL("https://api.figshare.com/v2/")
                ),
                9782777,
                2,
            ),
            MaybeExists(
                ImmutableFigshareDataset(
                    FigshareInstallation(
                        URL("https://figshare.com/"),
                        URL("https://api.figshare.com/v2/"),
                    ),
                    9782777,
                    2,
                )
            ),
        ),
        # Non existent things
        (
            FigshareDataset(
                FigshareInstallation(
                    URL("https://figshare.com/"), URL("https://api.figshare.com/v2/")
                ),
                97827778384384634634634863463434343,
                None,
            ),
            DoesNotExist(
                ImmutableFigshareDataset,
                "Article ID 97827778384384634634634863463434343 not found on figshare installation https://figshare.com/",
            ),
        ),
    ),
)
async def test_immutable_figshare(question, expected):
    ifs = ImmutableFigshareResolver()
    assert await ifs.resolve(question) == expected



import hashlib
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from repoproviders.fetchers import fetch
from repoproviders.resolvers import resolve
from repoproviders.resolvers.base import Exists


@pytest.mark.parametrize(
    ("questions", "md5tree"),
    [
        (
            (
                "doi:10.7910/DVN/TJCLKP",
                "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/TJCLKP",
            ),
            {
                "data/primary/primary-data.zip": "a8f6fc3fc58f503cd48e23fa8b088694",
                "data/2023-01-03.tsv": "6fd497bf13dab9a06fe737ebc22f1917",
                "code/language.py": "9d61582bcf497c83bbd1ed0eed3c772e",
            },
        ),
        (
            (
                "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ",
                "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ",
                "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            ),
            {
                "ArchaeoGLOBE-master/analysis/figures/1_response_distribution.png": "243c6a3dd66bc3c84102829b277ef333",
                "ArchaeoGLOBE-master/analysis/figures/2_trends_map_knowledge.png": "2ace6ae9d470dda6cf2f9f9a6588171a",
                "ArchaeoGLOBE-master/analysis/figures/3_trends_global.png": "63ccd0a7b2d20440cd8f418d4ee88c4d",
                "ArchaeoGLOBE-master/analysis/figures/4_consensus_transitions.png": "facfaedabeac77c4496d4b9e962a917f",
                "ArchaeoGLOBE-master/analysis/figures/5_ArchaeoGLOBE_HYDE_comparison.png": "8e002e4d50f179fc1808f562b1353588",
                "ArchaeoGLOBE-master/apt.txt": "b4224032da6c71d48f46c9b78fc6ed77",
                "ArchaeoGLOBE-master/analysis/archaeoglobe.pdf": "f575be4790efc963ef1bd40d097cc06d",
                "ArchaeoGLOBE-master/analysis/archaeoglobe.Rmd": "f37d5f7993fde9ebd64d16b20fc22905",
                "ArchaeoGLOBE-master/ArchaeoGLOBE.Rproj": "d0250e7918993bab1e707358fe5633e0",
                "ArchaeoGLOBE-master/CONDUCT.md": "f87ef290340322089c32b4e573d8f1e8",
                "ArchaeoGLOBE-master/.circleci/config.yml": "6eaa54073a682b3195d8fab3a9dd8344",
                "ArchaeoGLOBE-master/CONTRIBUTING.md": "b3a6abfc749dd155a3049f94a855bf9f",
                "ArchaeoGLOBE-master/DESCRIPTION": "745ef979494999e483987de72c0adfbd",
                "ArchaeoGLOBE-master/dockerfile": "aedce68e5a7d6e79cbb24c9cffeae593",
                "ArchaeoGLOBE-master/.binder/Dockerfile": "7564a41246ba99b60144afb1d3b6d7de",
                "ArchaeoGLOBE-master/.gitignore": "62c1482e4febbd35dc02fb7e2a31246b",
                "ArchaeoGLOBE-master/analysis/data/derived-data/hyde_crop_prop.RDS": "2aea7748b5586923b0de9d13af58e59d",
                "ArchaeoGLOBE-master/analysis/data/derived-data/kk_anthro_prop.RDS": "145a9e5dd2c95625626a720b52178b70",
                "ArchaeoGLOBE-master/LICENSE.md": "3aa9d41a92a57944bd4590e004898445",
                "ArchaeoGLOBE-master/analysis/data/derived-data/placeholder": "d41d8cd98f00b204e9800998ecf8427e",
                "ArchaeoGLOBE-master/.Rbuildignore": "df15e4fed49abd685b536fef4472b01f",
                "ArchaeoGLOBE-master/README.md": "0b0faabe580c4d76a0e0d64a4f54bca4",
                "ArchaeoGLOBE-master/analysis/data/derived-data/README.md": "547fd1a6e874f6178b1cf525b5b9ae72",
                "ArchaeoGLOBE-master/analysis/figures/S1_FHG_consensus.png": "d2584352e5442b33e4b23e361ca70fe1",
                "ArchaeoGLOBE-master/analysis/figures/S2_EXAG_consensus.png": "513eddfdad01fd01a20263a55ca6dbe3",
                "ArchaeoGLOBE-master/analysis/figures/S3_INAG_consensus.png": "b16ba0ecd21b326f873209a7e55a8deb",
                "ArchaeoGLOBE-master/analysis/figures/S4_PAS_consensus.png": "05695f9412337a00c1cb6d1757d0ec5c",
                "ArchaeoGLOBE-master/analysis/figures/S5_URBAN_consensus.png": "10119f7495d3b8e7ad7f8a0770574f15",
                "ArchaeoGLOBE-master/analysis/figures/S6_trends_map_landuse.png": "b1db7c97f39ccfc3a9e094c3e6307af0",
                "ArchaeoGLOBE-master/analysis/figures/S7_ArchaeoGLOBE_KK10_comparison.png": "30341748324f5f66acadb34c114c3e9d",
            },
        ),
    ],
)
async def test_fetch(questions: list[str], md5tree: dict[str, str]):
    for question in questions:
        with TemporaryDirectory() as d:
            output_dir = Path(d)
            answers = await resolve(question, True)

            assert answers is not None
            assert isinstance(answers[-1], Exists)

            await fetch(answers[-1].repo, output_dir)

            # Verify md5 sum of the files we expect to find
            # We are using md5 instead of something more secure because that is what
            # dataverse itself uses
            for subpath, expected_sha in md5tree.items():
                with open(output_dir / subpath, "rb") as f:
                    h = hashlib.md5()
                    h.update(f.read())
                    assert h.hexdigest() == expected_sha



import asyncio
import re

from aiohttp import ClientSession
from yarl import URL

from .base import DoesNotExist, Exists, MaybeExists
from .repos import GistURL, Git, GitHubPR, GitHubURL, GitLabURL, ImmutableGit


class GitHubResolver:
    async def resolve(
        self, question: GitHubURL
    ) -> MaybeExists[Git] | MaybeExists[GitHubPR] | None:
        url = question.url
        # Split the URL into parts, discarding empty parts to account for leading and trailing slashes
        parts = [p for p in url.path.split("/") if p.strip() != ""]
        if len(parts) == 2:
            # Handle <user|org>/<repo>
            # Reconstruct the URL so we normalize any
            return MaybeExists(
                Git(repo=str(url.with_path(f"{parts[0]}/{parts[1]}")), ref="HEAD")
            )
        elif len(parts) >= 4 and parts[2] in ("tree", "blob"):
            # Handle <user|org>/<repo>/<tree|blob>/<ref>(/<possible-path>)
            # Note: We ignore any paths specified here, as we only care about the repo
            return MaybeExists(
                Git(repo=str(url.with_path(f"{parts[0]}/{parts[1]}")), ref=parts[3])
            )
        elif len(parts) == 4 and parts[2] == "pull" and parts[3].isdigit():
            # Resolve pull requests to the branch their head ref points to
            return MaybeExists(GitHubPR(question.installation, question.url))
        else:
            # This is not actually a valid GitHub URL we support
            return None


class GitHubPRResolver:
    async def resolve(
        self, question: GitHubPR
    ) -> MaybeExists[Git] | DoesNotExist[GitHubPR] | None:
        parts = [p for p in question.url.path.split("/") if p.strip() != ""]
        org = parts[0]
        repo = parts[1]
        pull_req_id = int(parts[3])
        async with ClientSession() as session:
            # FIXME: Support enterprise github installations
            # FIXME: Add authentication support
            api_url = (
                URL("https://api.github.com/repos")
                / org
                / repo
                / "pulls"
                / str(pull_req_id)
            )
            resp = await session.get(api_url)

            data = await resp.json()
            if resp.status == 404:
                return DoesNotExist(
                    GitHubPR, f"PR {pull_req_id} does not exist at {question.url}"
                )
            repo_url = data["head"]["repo"]["html_url"]
            ref = data["head"]["ref"]
            return MaybeExists(Git(repo_url, ref))


class GistResolver:
    async def resolve(self, question: GistURL) -> MaybeExists[Git] | None:
        url = question.url
        # Split the URL into parts, discarding empty parts to account for leading and trailing slashes
        parts = [p for p in url.path.split("/") if p.strip() != ""]

        if len(parts) == 2:
            # Handle user/gist-id
            return MaybeExists[Git](repo=Git(str(question.url), "HEAD"))

        # FIXME: Handle ways to specify individual refs in Gist.
        # I can't seem to find a
        return None


class GitLabResolver:
    async def resolve(self, question: GitLabURL) -> MaybeExists[Git] | None:
        url = question.url
        # Split the URL into parts, discarding empty parts to account for leading and trailing slashes
        parts = [p for p in url.path.split("/") if p.strip() != ""]
        if len(parts) in (2, 3):
            # Handle <user|org>/<repo> as well as <user|org>/<namespace>/<repo>
            # Reconstruct the URL so we normalize any
            return MaybeExists(
                # Clear out the URL to remove query params & fragments
                Git(repo=str(url.with_query(None).with_fragment(None)), ref="HEAD")
            )
        elif "-" in parts:
            dash_index = parts.index("-")
            if len(parts) == dash_index + 1:
                # The dash is the last part of the URL, which isn't something we recognize to be anything
                return None
            if not parts[dash_index + 1] in ("tree", "blob"):
                # GitLab has dashes in lots of URLs, we only care about tree and blob ones
                return None
            return MaybeExists(
                Git(
                    str(url.with_path("/".join(parts[0:dash_index]))),
                    parts[dash_index + 2],
                )
            )
        else:
            # This is not actually a valid GitLab URL we support
            return None


class ImmutableGitResolver:
    async def resolve(
        self, question: Git
    ) -> (
        Exists[ImmutableGit]
        | MaybeExists[ImmutableGit]
        | DoesNotExist[ImmutableGit]
        | None
    ):
        command = ["git", "ls-remote", "--", question.repo, question.ref]
        proc = await asyncio.create_subprocess_exec(
            *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = [s.decode().strip() for s in await proc.communicate()]
        retcode = await proc.wait()
        if retcode:
            # `git` may follow redirects here, so the repo we pass may not always be the repo we
            # get back. So we loosely check for a 'not found' message.
            if re.search(r"fatal: repository '(.+)' not found", stderr, re.MULTILINE):
                return DoesNotExist(
                    ImmutableGit, f"Could not access git repository at {question.repo}"
                )

            # If it's another error, let's raise it directly
            raise RuntimeError(
                f"Unable to run git ls-remote to resolve {question}: {stderr}"
            )

        if stdout == "":
            # The remote repo exists, and we can talk to it. But no *ref* with the given name
            # exists. We check if this looks like a commit sha, and if it is, assume it exists.
            # FIXME: Decide if we just do this check *before* trying ls-remotes? Faster, but it means
            # we can't guarantee that the repo itself exists
            if re.match(r"[0-9a-f]{40}", question.ref):
                resolved_ref = question.ref
                return MaybeExists(ImmutableGit(question.repo, resolved_ref))
            else:
                return DoesNotExist(
                    ImmutableGit, f"No ref {question.ref} found in repo {question.repo}"
                )
        else:
            resolved_ref = stdout.split("\t", 1)[0]

            return Exists(ImmutableGit(question.repo, resolved_ref))


class GitUrlResolver:
    """
    Resolves raw git URLs

    URL structure is inspired by what `pip` supports: https://pip.pypa.io/en/stable/topics/vcs-support/#git
    """

    async def resolve(self, question: URL) -> MaybeExists[Git] | None:
        # List of supported protocols is from https://pip.pypa.io/en/stable/topics/vcs-support/#git
        if question.scheme not in (
            "git+https",
            "git+ssh",
            "git",
            "git+file",
            "git+http",
            "git+git",
        ):
            return None

        repo = question.with_scheme(question.scheme.replace("git+", ""))

        if "@" in question.path:
            parts = question.path.split("@", 1)
            ref = parts[1]
            repo = repo.with_path(parts[0])
        else:
            ref = "HEAD"

        return MaybeExists(Git(str(repo), ref))



import pytest
from yarl import URL

from repoproviders.resolvers.base import DoesNotExist, MaybeExists
from repoproviders.resolvers.doi import ZenodoResolver
from repoproviders.resolvers.repos import ZenodoDataset, ZenodoURL


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # A valid Zenodo URL that isn't actually a dataset
        (
            ZenodoURL(URL("https://zenodo.org"), URL("https://zenodo.org/communities")),
            None,
        ),
        # Simple /record and /records
        (
            ZenodoURL(
                URL("https://zenodo.org"), URL("https://zenodo.org/record/3232985")
            ),
            MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), "3232985")),
        ),
        (
            ZenodoURL(
                URL("https://zenodo.org"), URL("https://zenodo.org/records/3232985")
            ),
            MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), "3232985")),
        ),
        # Note we normalize output to have the HTTPS URL, even if we're passed in the HTTP URL
        (
            ZenodoURL(
                URL("https://zenodo.org"), URL("http://zenodo.org/record/3232985")
            ),
            MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), "3232985")),
        ),
        (
            ZenodoURL(
                URL("https://zenodo.org"), URL("https://zenodo.org/records/3232985")
            ),
            MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), "3232985")),
        ),
        # A non-zenodo.org URL
        (
            ZenodoURL(
                URL("https://data.caltech.edu"),
                URL("https://data.caltech.edu/records/996aw-mf266"),
            ),
            MaybeExists(ZenodoDataset(URL("https://data.caltech.edu/"), "996aw-mf266")),
        ),
        # A doi reference
        (
            ZenodoURL(
                URL("https://zenodo.org"),
                URL("https://zenodo.org/doi/10.5281/zenodo.805993"),
            ),
            MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), recordId="14007206")),
        ),
        # A doi reference to a bad doi
        (
            ZenodoURL(
                URL("https://zenodo.org"),
                URL("https://zenodo.org/doi/10.5281/zdo.805993"),
            ),
            DoesNotExist(
                ZenodoDataset,
                "https://zenodo.org/doi/10.5281/zdo.805993 is not a valid Zenodo DOI URL",
            ),
        ),
    ),
)
async def test_zenodo(url, expected):
    zr = ZenodoResolver()
    assert await zr.resolve(url) == expected



import pytest
from yarl import URL

from repoproviders.resolvers import resolve
from repoproviders.resolvers.base import DoesNotExist, Exists, MaybeExists
from repoproviders.resolvers.repos import (
    DataverseDataset,
    DataverseURL,
    Doi,
    FigshareDataset,
    FigshareInstallation,
    FigshareURL,
    GistURL,
    Git,
    GitHubPR,
    GitHubURL,
    GitLabURL,
    ImmutableFigshareDataset,
    ImmutableGit,
    ZenodoDataset,
    ZenodoURL,
)


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", []),
        # GitHub URLs that are not repos, but are still GitHub URLs
        (
            "https://github.com/pyOpenSci",
            [
                MaybeExists(
                    GitHubURL(
                        URL("https://github.com"), URL("https://github.com/pyOpenSci")
                    )
                )
            ],
        ),
        (
            "https://github.com/yuvipanda/repoproviders/actions/runs/12552733471/job/34999118812",
            [
                MaybeExists(
                    GitHubURL(
                        URL("https://github.com"),
                        URL(
                            "https://github.com/yuvipanda/repoproviders/actions/runs/12552733471/job/34999118812"
                        ),
                    )
                )
            ],
        ),
        (
            "https://github.com/yuvipanda/repoproviders/settings",
            [
                MaybeExists(
                    GitHubURL(
                        URL("https://github.com"),
                        URL("https://github.com/yuvipanda/repoproviders/settings"),
                    )
                )
            ],
        ),
        (
            "https://github.com/jupyter/docker-stacks/pull/2194",
            [
                MaybeExists(
                    GitHubURL(
                        URL("https://github.com"),
                        URL("https://github.com/jupyter/docker-stacks/pull/2194"),
                    )
                )
            ],
        ),
        # Simple github repo URL that are actual repos
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/pyOpenSci/pyos-package-template"),
            ),
            [
                MaybeExists(
                    Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD")
                )
            ],
        ),
        # Trailing slash normalized?
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/pyOpenSci/pyos-package-template/"),
            ),
            [
                MaybeExists(
                    Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD")
                )
            ],
        ),
        # blobs and tree
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/pyOpenSci/pyos-package-template/tree/main/includes/licenses"
                ),
            ),
            [
                MaybeExists(
                    Git("https://github.com/pyOpenSci/pyos-package-template", "main")
                )
            ],
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/pyOpenSci/pyos-package-template/tree/original-cookie/docs"
                ),
            ),
            [
                MaybeExists(
                    Git(
                        "https://github.com/pyOpenSci/pyos-package-template",
                        "original-cookie",
                    )
                )
            ],
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/pyOpenSci/pyos-package-template/blob/b912433bfae541972c83529359f4181ef0fe9b67/README.md"
                ),
            ),
            [
                MaybeExists(
                    Git(
                        "https://github.com/pyOpenSci/pyos-package-template",
                        ref="b912433bfae541972c83529359f4181ef0fe9b67",
                    )
                )
            ],
        ),
        # Random URL, not a git repo
        (
            Git("https://example.com/something", "HEAD"),
            [
                DoesNotExist(
                    ImmutableGit,
                    "Could not access git repository at https://example.com/something",
                )
            ],
        ),
        # Resolve a tag
        (
            Git("https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "0.8.0"),
            [
                Exists(
                    ImmutableGit(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                    )
                )
            ],
        ),
        # Resolve a commit we know exists, although this isn't verified
        (
            Git(
                "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            ),
            [
                MaybeExists(
                    ImmutableGit(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                    )
                )
            ],
        ),
        (
            Git(
                "https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "does-not-exist"
            ),
            [
                DoesNotExist(
                    ImmutableGit,
                    "No ref does-not-exist found in repo https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                )
            ],
        ),
    ),
)
async def test_resolve(url, expected):
    assert await resolve(url, False) == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        ("https://example.com/something", []),
        # doi schema'd URI
        (
            "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        )
                    )
                )
            ],
        ),
        # handle schema'd URI
        (
            "hdl:11529/10016",
            [
                Exists(
                    Doi(
                        URL(
                            "https://data.cimmyt.org/dataset.xhtml?persistentId=hdl:11529/10016"
                        )
                    )
                )
            ],
        ),
        # For convenience, we do accept DOIs without a scheme
        (
            "10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        )
                    )
                )
            ],
        ),
        # But not handles without a scheme
        ("11529/10016", []),
        # Three DOI resolution URLs
        (
            "https://doi.org/10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        )
                    )
                )
            ],
        ),
        (
            "https://www.doi.org/10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        )
                    )
                )
            ],
        ),
        (
            "https://hdl.handle.net/10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        )
                    )
                )
            ],
        ),
    ),
)
async def test_norecurse(url, expected):
    assert await resolve(URL(url), False) == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # doi schema'd URI
        (
            "doi:10.7910/DVN/6ZXAGT/3YRRYJ",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        )
                    )
                ),
                MaybeExists(
                    DataverseURL(
                        URL("https://dataverse.harvard.edu"),
                        URL(
                            "https://dataverse.harvard.edu/file.xhtml?persistentId=doi:10.7910/DVN/6ZXAGT/3YRRYJ"
                        ),
                    ),
                ),
                Exists(
                    DataverseDataset(
                        URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
                    )
                ),
            ],
        ),
        # handle schema'd URI
        (
            "hdl:11529/10016",
            [
                Exists(
                    Doi(
                        URL(
                            "https://data.cimmyt.org/dataset.xhtml?persistentId=hdl:11529/10016"
                        )
                    )
                ),
                MaybeExists(
                    DataverseURL(
                        URL("https://data.cimmyt.org"),
                        URL(
                            "https://data.cimmyt.org/dataset.xhtml?persistentId=hdl:11529/10016"
                        ),
                    ),
                ),
                Exists(
                    DataverseDataset(URL("https://data.cimmyt.org"), "hdl:11529/10016")
                ),
            ],
        ),
        # For convenience, we do accept DOIs without a scheme
        (
            "10.7910/DVN/6ZXAGT",
            [
                Exists(
                    Doi(
                        URL(
                            "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/6ZXAGT"
                        )
                    )
                ),
                MaybeExists(
                    DataverseURL(
                        URL("https://dataverse.harvard.edu"),
                        URL(
                            "https://dataverse.harvard.edu/citation?persistentId=doi:10.7910/DVN/6ZXAGT"
                        ),
                    ),
                ),
                Exists(
                    DataverseDataset(
                        URL("https://dataverse.harvard.edu"), "doi:10.7910/DVN/6ZXAGT"
                    )
                ),
            ],
        ),
        # Something that's only a DOI, and won't resolve further
        (
            "10.1126/science.aar3646",
            [Exists(Doi(URL("https://www.science.org/doi/10.1126/science.aar3646")))],
        ),
        # GitHub URLs that recurse into ImmutableGit
        (
            "https://github.com/jupyterhub/zero-to-jupyterhub-k8s/tree/f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            [
                MaybeExists(
                    GitHubURL(
                        URL("https://github.com"),
                        URL(
                            "https://github.com/jupyterhub/zero-to-jupyterhub-k8s/tree/f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603"
                        ),
                    )
                ),
                MaybeExists(
                    Git(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                    )
                ),
                MaybeExists(
                    ImmutableGit(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                    )
                ),
            ],
        ),
        (
            "https://github.com/jupyterhub/zero-to-jupyterhub-k8s/tree/0.8.0",
            [
                MaybeExists(
                    GitHubURL(
                        URL("https://github.com"),
                        URL(
                            "https://github.com/jupyterhub/zero-to-jupyterhub-k8s/tree/0.8.0"
                        ),
                    )
                ),
                MaybeExists(
                    Git(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "0.8.0",
                    )
                ),
                Exists(
                    ImmutableGit(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                    )
                ),
            ],
        ),
        # Git URLs that recurse into ImmutableGit
        (
            "git+https://github.com/jupyterhub/zero-to-jupyterhub-k8s@f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            [
                MaybeExists(
                    Git(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                    )
                ),
                MaybeExists(
                    ImmutableGit(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                    )
                ),
            ],
        ),
        (
            "git+https://github.com/jupyterhub/zero-to-jupyterhub-k8s@0.8.0",
            [
                MaybeExists(
                    Git(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "0.8.0",
                    )
                ),
                Exists(
                    ImmutableGit(
                        "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                        "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                    )
                ),
            ],
        ),
        (
            "10.5281/zenodo.3232985",
            [
                Exists(Doi(URL("https://zenodo.org/record/3232985"))),
                MaybeExists(
                    ZenodoURL(
                        URL("https://zenodo.org"),
                        URL("https://zenodo.org/record/3232985"),
                    )
                ),
                MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), "3232985")),
            ],
        ),
        (
            "https://doi.org/10.6084/m9.figshare.9782777.v3",
            [
                Exists(
                    Doi(
                        URL(
                            "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777/3"
                        )
                    )
                ),
                MaybeExists(
                    FigshareURL(
                        FigshareInstallation(
                            URL("https://figshare.com/"),
                            URL("https://api.figshare.com/v2/"),
                        ),
                        URL(
                            "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777/3"
                        ),
                    )
                ),
                MaybeExists(
                    FigshareDataset(
                        FigshareInstallation(
                            URL("https://figshare.com/"),
                            URL("https://api.figshare.com/v2/"),
                        ),
                        9782777,
                        3,
                    )
                ),
                MaybeExists(
                    ImmutableFigshareDataset(
                        FigshareInstallation(
                            URL("https://figshare.com/"),
                            URL("https://api.figshare.com/v2/"),
                        ),
                        9782777,
                        3,
                    )
                ),
            ],
        ),
        (
            "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777",
            [
                MaybeExists(
                    FigshareURL(
                        FigshareInstallation(
                            URL("https://figshare.com/"),
                            URL("https://api.figshare.com/v2/"),
                        ),
                        URL(
                            "https://figshare.com/articles/Binder-ready_openSenseMap_Analysis/9782777"
                        ),
                    )
                ),
                MaybeExists(
                    FigshareDataset(
                        FigshareInstallation(
                            URL("https://figshare.com/"),
                            URL("https://api.figshare.com/v2/"),
                        ),
                        9782777,
                        None,
                    )
                ),
                Exists(
                    ImmutableFigshareDataset(
                        FigshareInstallation(
                            URL("https://figshare.com/"),
                            URL("https://api.figshare.com/v2/"),
                        ),
                        9782777,
                        3,
                    )
                ),
            ],
        ),
        # A zenodo doi that uses the /doi redirect
        (
            "https://doi.org/10.5281/zenodo.805993",
            [
                Exists(Doi(url=URL("https://zenodo.org/doi/10.5281/zenodo.805993"))),
                MaybeExists(
                    ZenodoURL(
                        URL("https://zenodo.org"),
                        URL("https://zenodo.org/doi/10.5281/zenodo.805993"),
                    )
                ),
                MaybeExists(ZenodoDataset(URL("https://zenodo.org/"), "14007206")),
            ],
        ),
        # A bare git URL, that we'll have to have guessed
        (
            # Using this as HEAD hasn't changed in 16 years
            "https://git.kernel.org/pub/scm/fs/fat/fatattr/fatattr.git/",
            [
                Exists(
                    repo=Git(
                        "https://git.kernel.org/pub/scm/fs/fat/fatattr/fatattr.git/",
                        ref="HEAD",
                    )
                ),
                Exists(
                    repo=ImmutableGit(
                        "https://git.kernel.org/pub/scm/fs/fat/fatattr/fatattr.git/",
                        "3df926a6a9ad5ea02c9f63157a0588125f046441",
                    )
                ),
            ],
        ),
        # A dataverse URL that does *not* exist in our list of well known dataverse installations
        (
            "https://demo.dataverse.org/dataset.xhtml?persistentId=doi:10.70122/FK2/MBQA9G",
            [
                MaybeExists(
                    DataverseURL(
                        URL("https://demo.dataverse.org"),
                        URL(
                            "https://demo.dataverse.org/dataset.xhtml?persistentId=doi:10.70122/FK2/MBQA9G"
                        ),
                    )
                ),
                Exists(
                    DataverseDataset(
                        URL("https://demo.dataverse.org"), "doi:10.70122/FK2/MBQA9G"
                    )
                ),
            ],
        ),
        # A GitLab URL on GitLab
        (
            "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder/-/blob/b2c44e7f804dc1634681540582b1731e0393f69b/03_Same_Time_Loop/_01_controller_master.ipynb?ref_type=heads",
            [
                MaybeExists(
                    repo=GitLabURL(
                        installation=URL("https://gitlab.com"),
                        url=URL(
                            "https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder/-/blob/b2c44e7f804dc1634681540582b1731e0393f69b/03_Same_Time_Loop/_01_controller_master.ipynb?ref_type=heads"
                        ),
                    )
                ),
                MaybeExists(
                    repo=Git(
                        repo="https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder",
                        ref="b2c44e7f804dc1634681540582b1731e0393f69b",
                    )
                ),
                MaybeExists(
                    repo=ImmutableGit(
                        repo="https://gitlab.com/mosaik/examples/mosaik-tutorials-on-binder",
                        ref="b2c44e7f804dc1634681540582b1731e0393f69b",
                    )
                ),
            ],
        ),
        # A GitLab URL that isn't on gitlab.com
        (
            "https://gitlab.wikimedia.org/toolforge-repos/toolviews/-/tree/bb42ab4dc4ddf0712f83ec4add58005a3ae75de5/toolviews?ref_type=heads",
            [
                MaybeExists(
                    repo=GitLabURL(
                        installation=URL("https://gitlab.wikimedia.org/"),
                        url=URL(
                            "https://gitlab.wikimedia.org/toolforge-repos/toolviews/-/tree/bb42ab4dc4ddf0712f83ec4add58005a3ae75de5/toolviews?ref_type=heads"
                        ),
                    )
                ),
                MaybeExists(
                    repo=Git(
                        repo="https://gitlab.wikimedia.org/toolforge-repos/toolviews",
                        ref="bb42ab4dc4ddf0712f83ec4add58005a3ae75de5",
                    )
                ),
                MaybeExists(
                    repo=ImmutableGit(
                        repo="https://gitlab.wikimedia.org/toolforge-repos/toolviews",
                        ref="bb42ab4dc4ddf0712f83ec4add58005a3ae75de5",
                    )
                ),
            ],
        ),
        # A gist URL
        (
            "https://gist.github.com/JakeWharton/5423616",
            [
                MaybeExists(
                    GistURL(
                        installation=URL("https://gist.github.com"),
                        url=URL("https://gist.github.com/JakeWharton/5423616"),
                    )
                ),
                MaybeExists(
                    repo=Git(
                        repo="https://gist.github.com/JakeWharton/5423616", ref="HEAD"
                    )
                ),
                Exists(
                    repo=ImmutableGit(
                        repo="https://gist.github.com/JakeWharton/5423616",
                        ref="76d24e01654211591b7bea8ae4557f6ff5283343",
                    )
                ),
            ],
        ),
        # A github PR that has been merged and branch deleted
        (
            "https://github.com/yuvipanda/repoproviders/pull/1",
            [
                MaybeExists(
                    repo=GitHubURL(
                        installation=URL("https://github.com"),
                        url=URL("https://github.com/yuvipanda/repoproviders/pull/1"),
                    )
                ),
                MaybeExists(
                    repo=GitHubPR(
                        installation=URL("https://github.com"),
                        url=URL("https://github.com/yuvipanda/repoproviders/pull/1"),
                    )
                ),
                MaybeExists(
                    repo=Git(
                        repo="https://github.com/yuvipanda/repoproviders", ref="types-1"
                    )
                ),
                DoesNotExist(
                    kind=ImmutableGit,
                    message="No ref types-1 found in repo https://github.com/yuvipanda/repoproviders",
                ),
            ],
        ),
    ),
)
async def test_recurse(url, expected):
    assert await resolve(URL(url), True) == expected



import pytest
from yarl import URL

from repoproviders.resolvers.base import DoesNotExist, MaybeExists
from repoproviders.resolvers.git import Git, GitHubPRResolver, GitHubResolver
from repoproviders.resolvers.repos import GitHubPR, GitHubURL


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # GitHub URLs that are not repos
        (
            GitHubURL(URL("https://github.com"), URL("https://github.com/pyOpenSci")),
            None,
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/yuvipanda/repoproviders/actions/runs/12552733471/job/34999118812"
                ),
            ),
            None,
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/yuvipanda/repoproviders/settings"),
            ),
            None,
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/jupyter/docker-stacks/pull/2194"),
            ),
            MaybeExists(
                GitHubPR(
                    URL("https://github.com"),
                    URL("https://github.com/jupyter/docker-stacks/pull/2194"),
                )
            ),
        ),
        # Simple github repo URL
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/pyOpenSci/pyos-package-template"),
            ),
            MaybeExists(
                Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD")
            ),
        ),
        # Trailing slash normalized?
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/pyOpenSci/pyos-package-template/"),
            ),
            MaybeExists(
                Git("https://github.com/pyOpenSci/pyos-package-template", "HEAD")
            ),
        ),
        # blobs and tree
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/pyOpenSci/pyos-package-template/tree/main/includes/licenses"
                ),
            ),
            MaybeExists(
                Git("https://github.com/pyOpenSci/pyos-package-template", "main")
            ),
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/pyOpenSci/pyos-package-template/tree/original-cookie/docs"
                ),
            ),
            MaybeExists(
                Git(
                    "https://github.com/pyOpenSci/pyos-package-template",
                    "original-cookie",
                )
            ),
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL(
                    "https://github.com/pyOpenSci/pyos-package-template/blob/b912433bfae541972c83529359f4181ef0fe9b67/README.md"
                ),
            ),
            MaybeExists(
                Git(
                    "https://github.com/pyOpenSci/pyos-package-template",
                    ref="b912433bfae541972c83529359f4181ef0fe9b67",
                )
            ),
        ),
        (
            GitHubURL(
                URL("https://github.com"),
                URL("https://github.com/yuvipanda/does-not-exist-e43"),
            ),
            MaybeExists(
                Git(repo="https://github.com/yuvipanda/does-not-exist-e43", ref="HEAD")
            ),
        ),
    ),
)
async def test_github(url, expected):
    gh = GitHubResolver()
    assert await gh.resolve(url) == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        (
            GitHubPR(
                URL("https://github.com"),
                URL("https://github.com/jupyter/docker-stacks/pull/2194"),
            ),
            MaybeExists(
                Git(
                    "https://github.com/mathbunnyru/docker-stacks",
                    "update_oracledb_version",
                )
            ),
        ),
        (
            GitHubPR(
                URL("https://github.com"),
                URL("https://github.com/jupyter/docker-stacks/pull/219400000000"),
            ),
            DoesNotExist(
                GitHubPR,
                "PR 219400000000 does not exist at https://github.com/jupyter/docker-stacks/pull/219400000000",
            ),
        ),
    ),
)
async def test_github_pr(url, expected):
    gh_pr = GitHubPRResolver()
    assert await gh_pr.resolve(url) == expected



import pytest

from repoproviders.resolvers.base import DoesNotExist, Exists, MaybeExists
from repoproviders.resolvers.git import Git, ImmutableGit, ImmutableGitResolver


@pytest.mark.parametrize(
    ("question", "expected"),
    (
        # Random URL, not a git repo
        (
            Git("https://example.com/something", "HEAD"),
            DoesNotExist(
                ImmutableGit,
                "Could not access git repository at https://example.com/something",
            ),
        ),
        # Resolve a tag
        (
            Git("https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "0.8.0"),
            Exists(
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "ada2170a2181ae1760d85eab74e5264d0c6bb67f",
                )
            ),
        ),
        # Resolve a commit we know exists, although this isn't verified
        (
            Git(
                "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
            ),
            MaybeExists(
                ImmutableGit(
                    "https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
                    "f7f3ff6d1bf708bdc12e5f10e18b2a90a4795603",
                )
            ),
        ),
        # Repo doesn't exist
        (
            Git(repo="https://github.com/yuvipanda/does-not-exist-e43", ref="HEAD"),
            DoesNotExist(
                ImmutableGit,
                "Could not access git repository at https://github.com/yuvipanda/does-not-exist-e43",
            ),
        ),
        # Ref doesn't exist
        (
            Git(
                "https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "does-not-exist"
            ),
            DoesNotExist(
                ImmutableGit,
                "No ref does-not-exist found in repo https://github.com/jupyterhub/zero-to-jupyterhub-k8s",
            ),
        ),
    ),
)
async def test_immutable_git(question, expected):
    ig = ImmutableGitResolver()
    assert await ig.resolve(question) == expected


async def test_immutable_git_HEAD():
    """
    Extra test to test resolving HEAD, making sure it's the same as resolving "main".

    This can't be fixtured because HEAD and `mail` will keep mocing
    """
    ig = ImmutableGitResolver()
    assert (
        await ig.resolve(
            Git("https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "HEAD")
        )
    ) == (
        await ig.resolve(
            Git("https://github.com/jupyterhub/zero-to-jupyterhub-k8s", "main")
        )
    )
